-------------------------------------------------------------------------------
-- Title      : package
-- Project    : 
-------------------------------------------------------------------------------
-- File       : package.vhd
-- Author     : Asbjrn Djupdal  <asbjoern@djupdal.org>
--            : Kjetil Aamodt
--            : Ola Martin Tiseth Stoevneng
-- Company    : 
-- Last update: 2013/12/10
-- Platform   : Spartan6
-------------------------------------------------------------------------------
-- Description: Constants and component declarations
-------------------------------------------------------------------------------
-- Revisions  :
-- Date        Version  Author    Description
-- 2013/12/10  3.0      stoevneng
-- 2005/05/23  2.0      aamodt 
-- 2003/01/17  1.1      djupdal
-- 2002/10/20  1.0      djupdal   Created
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use IEEE.numeric_std.all;
use work.funct_package.all;

package sblock_package is

  -- constants set by preprosessor

  constant COORD_SIZE_X : integer := 4 ;
  constant COORD_SIZE_Y : integer := 3 ;

  -- constants

  constant ROWS    : integer := 2 ** COORD_SIZE_Y;
  constant COLUMNS : integer := 2 ** COORD_SIZE_X;

  constant RULE_SIZE         : integer := 49;
  constant RULE_NBR_BUS_SIZE : integer := 8;
  constant RULES_IN_SET      : integer := 8;

  constant LUT_SIZE  : integer := 32;
  constant TYPE_SIZE : integer := 5;

  constant ADDR_BUS_SIZE : integer := COORD_SIZE_X + COORD_SIZE_Y - 2;
--Kaa
  -- for memory contain data generated on run-steps
  -- address bus can be changed to expand the numbers of values that can be stored
  constant RUN_STEP_ADDR_BUS_SIZE : integer := 13;
  constant RUN_STEP_DATA_BUS_SIZE : integer := 16;

  -- for BRAM used to store fired rules for many dev-steps
  -- can NOT be changed, BRAM not dynamic
  constant RULEVECTOR_ADDR_BUS_SIZE : integer := 8;
  constant RULEVECTOR_DATA_BUS_SIZE : integer := 256;

  -- for BRAM used to store fired rules for one dev-step
  constant USEDRULES_ADDR_BUS_SIZE : integer := 9;
  constant USEDRULES_DATA_BUS_SIZE : integer := 16;

  -- for fitness result storage
  constant FITNESS_RESULT_SIZE  : integer := 4 * 32; 
  constant FITNESS_DATA_BUS_SIZE: integer := 32;
--Kaa

  constant SBM_CFG_SIZE : integer := 32;
  constant SBM_RDB_SIZE : integer := 16;

  constant TYPE_BUS_SIZE  : integer := TYPE_SIZE * 2;
  constant STATE_BUS_SIZE : integer := 2;

  constant READBACK_WORDS : integer := 
    ceil ((ROWS * COLUMNS) / SBM_RDB_SIZE, (ROWS * COLUMNS) mod SBM_RDB_SIZE);
  constant CONFIG_WORDS : integer := 
    ceil ((ROWS * COLUMNS) / SBM_CFG_SIZE, (ROWS * COLUMNS) mod SBM_CFG_SIZE);

  -----------------------------------------------------------------------------
  -- types

  type rule_set_t is array (RULES_IN_SET - 1 downto 0)
    of std_logic_vector(RULE_SIZE - 1 downto 0);

  type exec_result_t is array (RULES_IN_SET - 1 downto 0)
    of std_logic_vector(TYPE_SIZE - 1 downto 0);

  -----------------------------------------------------------------------------
  -- components from Xilinx library

  component SRL16E
    port (
      d   : in  std_logic;
      ce  : in  std_logic;
      a0  : in  std_logic;
      a1  : in  std_logic;
      a2  : in  std_logic;
      a3  : in  std_logic;
      q   : out std_logic;
      clk : in  std_logic);
  end component;

  component bufg
    port (
      i : in std_logic;
      o : out std_logic);
  end component;

  component ibufg
    port (
      i : in  std_logic;
      o : out std_logic);
  end component;

  component clkdll
    port (
      clkin  : in  std_logic;
      clkfb  : in  std_logic;
      rst    : in  std_logic;
      clk0   : out std_logic;
      clk90  : out std_logic;
      clk180 : out std_logic;
      clk270 : out std_logic;
      clk2x  : out std_logic;
      clkdv  : out std_logic;
      locked : out std_logic);
  end component;

  -----------------------------------------------------------------------------
  -- components part of this project

  component toplevel
    port (
      pciBusy   : in    std_logic;
      pciEmpty  : in    std_logic;
      pciRW     : out   std_logic;
      pciEnable : out   std_logic;
      pciData   : inout std_logic_vector(31 downto 0);
      debugled  : out   std_logic_vector(15 downto 0);
      clk       : in    std_logic;
      clk40     : in    std_logic;
      rst       : in    std_logic);
  end component;

  -----------------------------------------------------------------------------

  component counter
    generic (
      SIZE : integer);
    port (
      reset    : in  std_logic;
      count    : in  std_logic;
      count_to : in  unsigned(SIZE - 1 downto 0);
      zero     : out std_logic;
      finished : out std_logic;
      value    : out std_logic_vector(SIZE - 1 downto 0);
      clk      : in  std_logic);
  end component;

  component addr_gen
    port (
      x             : in  std_logic_vector(COORD_SIZE_X - 1 downto 0);
      y             : in  std_logic_vector(COORD_SIZE_Y - 1 downto 0);
      addr          : out std_logic_vector(ADDR_BUS_SIZE - 1 downto 0);
      sblock_number : out std_logic;
      port_select   : out std_logic);
  end component;

  component word_select
    generic (
      NUMBER_OF_WORDS : integer);
    port (
      reset_select  : in  std_logic;
      select_next   : in  std_logic;
      selected_word : out std_logic_vector(NUMBER_OF_WORDS - 1 downto 0);
      rst           : in  std_logic;
      clk           : in  std_logic);
  end component;

  -----------------------------------------------------------------------------

  component instr_mem
    port (
      addr         : in  std_logic_vector(7 downto 0);
      data_read    : out std_logic_vector(63 downto 0);
      data_write   : in  std_logic_vector(63 downto 0);
      write_enable : in  std_logic;
      stall        : in  std_logic;
      rst          : in  std_logic;
      clk          : in  std_logic);
  end component;

  component sblock
    port (
      east              : in  std_logic;
      south             : in  std_logic;
      north             : in  std_logic;
      west              : in  std_logic;
      conf_data_l       : in  std_logic;
      conf_data_h       : in  std_logic;
      conf_data_ff      : in  std_logic;
      config_lut_enable : in  std_logic;
      config_ff_enable  : in  std_logic;
      output            : out std_logic;
      run               : in  std_logic;
      clk               : in  std_logic);
  end component;

  component sbm_bram
    port (
      type_data_read_a     : out std_logic_vector(TYPE_BUS_SIZE - 1 downto 0);
      type_data_read_b     : out std_logic_vector(TYPE_BUS_SIZE - 1 downto 0);
      type_data_read_c     : out std_logic_vector(TYPE_BUS_SIZE - 1 downto 0);
      type_data_read_d     : out std_logic_vector(TYPE_BUS_SIZE - 1 downto 0);
      state_data_read_a    : out std_logic_vector(STATE_BUS_SIZE - 1 downto 0);
      state_data_read_b    : out std_logic_vector(STATE_BUS_SIZE - 1 downto 0);
      state_data_read_c    : out std_logic_vector(STATE_BUS_SIZE - 1 downto 0);
      state_data_read_d    : out std_logic_vector(STATE_BUS_SIZE - 1 downto 0);
      type_data_write_a    : in  std_logic_vector(TYPE_BUS_SIZE - 1 downto 0);
      type_data_write_b    : in  std_logic_vector(TYPE_BUS_SIZE - 1 downto 0);
      type_data_write_c    : in  std_logic_vector(TYPE_BUS_SIZE - 1 downto 0);
      type_data_write_d    : in  std_logic_vector(TYPE_BUS_SIZE - 1 downto 0);
      state_data_write_a   : in  std_logic_vector(STATE_BUS_SIZE - 1 downto 0);
      state_data_write_b   : in  std_logic_vector(STATE_BUS_SIZE - 1 downto 0);
      state_data_write_c   : in  std_logic_vector(STATE_BUS_SIZE - 1 downto 0);
      state_data_write_d   : in  std_logic_vector(STATE_BUS_SIZE - 1 downto 0);
      addr_a               : in  std_logic_vector(ADDR_BUS_SIZE - 1 downto 0);
      addr_b               : in  std_logic_vector(ADDR_BUS_SIZE - 1 downto 0);
      addr_c               : in  std_logic_vector(ADDR_BUS_SIZE - 1 downto 0);
      addr_d               : in  std_logic_vector(ADDR_BUS_SIZE - 1 downto 0);
      enable_read_a        : in  std_logic;
      enable_read_b        : in  std_logic;
      enable_read_c        : in  std_logic;
      enable_read_d        : in  std_logic;
      enable_type_write_a  : in  std_logic;
      enable_type_write_b  : in  std_logic;
      enable_type_write_c  : in  std_logic;
      enable_type_write_d  : in  std_logic;
      enable_state_write_a : in  std_logic;
      enable_state_write_b : in  std_logic;
      enable_state_write_c : in  std_logic;
      enable_state_write_d : in  std_logic;
      stall                : in  std_logic;
      clk                  : in  std_logic;
      rst                  : in  std_logic);
  end component;

  component bram_inferrer
    generic (
      addr_bits : integer;
      data_bits : integer);
    port (
      clk_a : in std_logic;
      clk_b : in std_logic;

      addr_a : in std_logic_vector(addr_bits - 1 downto 0);
      data_i_a : in std_logic_vector(data_bits - 1 downto 0);
      data_o_a : out std_logic_vector(data_bits - 1 downto 0);
      we_a : in std_logic;
      en_a : in std_logic;
      rst_a : in std_logic;

      addr_b : in std_logic_vector(addr_bits - 1 downto 0);
      data_i_b : in std_logic_vector(data_bits - 1 downto 0);
      data_o_b : out std_logic_vector(data_bits - 1 downto 0);
      we_b : in std_logic;
      en_b : in std_logic;
      rst_b : in std_logic);
  end component;

  component srl_inferer
    generic (
      size : integer;
      a_size : integer);
    port (
      d : in std_logic;
      ce : in std_logic;
      a : in std_logic_vector(a_size - 1 downto 0);
      q : out std_logic;
      clk : in std_logic);
  end component;


  component rule_executive
    port (
      north_type   : in  std_logic_vector(TYPE_SIZE - 1 downto 0);
      south_type   : in  std_logic_vector(TYPE_SIZE - 1 downto 0);
      east_type    : in  std_logic_vector(TYPE_SIZE - 1 downto 0);
      west_type    : in  std_logic_vector(TYPE_SIZE - 1 downto 0);
      center_type  : in  std_logic_vector(TYPE_SIZE - 1 downto 0);
      north_state  : in  std_logic;
      south_state  : in  std_logic;
      east_state   : in  std_logic;
      west_state   : in  std_logic;
      center_state : in  std_logic;
      rule         : in  std_logic_vector(RULE_SIZE - 1 downto 0);
      hit          : out std_logic;
      result_type  : out std_logic_vector(TYPE_SIZE - 1 downto 0);
      result_state : out std_logic;
      rst          : in  std_logic;
      clk          : in  std_logic);
  end component;

  component rule_select
    port (
      hits          : in  std_logic_vector(RULES_IN_SET - 1 downto 0);
      old_type      : in  std_logic_vector(TYPE_SIZE - 1 downto 0);
      old_state     : in  std_logic;
      results_type  : in  exec_result_t;
      results_state : in  std_logic_vector(RULES_IN_SET - 1 downto 0);
      result_type   : out std_logic_vector(TYPE_SIZE - 1 downto 0);
      result_state  : out std_logic;
--Kaa
      old_selected_rule: in std_logic_vector(7 downto 0);
      selected_rule    :out std_logic_vector(7 downto 0);
      ruleset          : in std_logic_vector(4 downto 0);
--Kaa
      rst           : in  std_logic;
      clk           : in  std_logic);
  end component;

  -----------------------------------------------------------------------------

  component com40
    port (
      send         : in    std_logic;
      ack_send     : out   std_logic;
      receive      : in    std_logic;
      ack_receive  : out   std_logic;
      data_send    : in    std_logic_vector(31 downto 0);
      data_receive : out   std_logic_vector(31 downto 0);
      pciBusy      : in    std_logic;
      pciEmpty     : in    std_logic;
      pciRW        : out   std_logic;
      pciEnable    : out   std_logic;
      pciData      : inout std_logic_vector(31 downto 0);
      rst          : in    std_logic;
      clk40        : in    std_logic);
  end component;

  component rule_storage
    port (
      ruleset        : out rule_set_t;
      cache_set_zero : in  std_logic;
      cache_next_set : in  std_logic;
      last_set       : out std_logic;
      store_rule     : in  std_logic;
      rule_number    : in  std_logic_vector(RULE_NBR_BUS_SIZE - 1 downto 0);
      rule_to_store  : in  std_logic_vector(RULE_SIZE - 1 downto 0);
      nbr_of_last_rule : in std_logic_vector(7 downto 0);
      rst                 : in std_logic;
      clk                 : in std_logic);
  end component;

  component lutconv
    port (
      index_a    : in  std_logic_vector(TYPE_SIZE - 1 downto 0);
      index_b    : in  std_logic_vector(TYPE_SIZE - 1 downto 0);
      lut_a_read : out std_logic_vector(LUT_SIZE - 1 downto 0);
      lut_b_read : out std_logic_vector(LUT_SIZE - 1 downto 0);
      lut_write  : in  std_logic_vector(LUT_SIZE - 1 downto 0);
      write_en   : in  std_logic;
      rst        : in  std_logic;
      clk        : in  std_logic);
  end component;

  component sblock_matrix
    port (
      databus_read        : out std_logic_vector(SBM_RDB_SIZE - 1 downto 0);
      output_select       : in  std_logic_vector(READBACK_WORDS - 1 downto 0);
      databus_lut_l_write : in  std_logic_vector(SBM_CFG_SIZE - 1 downto 0);
      databus_lut_h_write : in  std_logic_vector(SBM_CFG_SIZE - 1 downto 0);
      databus_ff_write    : in  std_logic_vector(SBM_CFG_SIZE - 1 downto 0);
      config_enable_lut   : in  std_logic_vector(CONFIG_WORDS - 1 downto 0);
      config_enable_ff    : in  std_logic_vector(CONFIG_WORDS - 1 downto 0);
      run_matrix          : in  std_logic;
      rst                 : in  std_logic;
      clk                 : in  std_logic);
  end component;

  component sbm_bram_mgr
    port (
      type_data_read_a0   : out std_logic_vector(TYPE_BUS_SIZE - 1 downto 0);
      type_data_read_b0   : out std_logic_vector(TYPE_BUS_SIZE - 1 downto 0);
      type_data_read_c0   : out std_logic_vector(TYPE_BUS_SIZE - 1 downto 0);
      type_data_read_d0   : out std_logic_vector(TYPE_BUS_SIZE - 1 downto 0);
      type_data_write_a0  : in  std_logic_vector(TYPE_BUS_SIZE - 1 downto 0);
      type_data_write_b0  : in  std_logic_vector(TYPE_BUS_SIZE - 1 downto 0);
      type_data_write_c0  : in  std_logic_vector(TYPE_BUS_SIZE - 1 downto 0);
      type_data_write_d0  : in  std_logic_vector(TYPE_BUS_SIZE - 1 downto 0);
      state_data_read_a0  : out std_logic_vector(STATE_BUS_SIZE - 1 downto 0);
      state_data_read_b0  : out std_logic_vector(STATE_BUS_SIZE - 1 downto 0);
      state_data_read_c0  : out std_logic_vector(STATE_BUS_SIZE - 1 downto 0);
      state_data_read_d0  : out std_logic_vector(STATE_BUS_SIZE - 1 downto 0);
      state_data_write_a0 : in  std_logic_vector(STATE_BUS_SIZE - 1 downto 0);
      state_data_write_b0 : in  std_logic_vector(STATE_BUS_SIZE - 1 downto 0);
      state_data_write_c0 : in  std_logic_vector(STATE_BUS_SIZE - 1 downto 0);
      state_data_write_d0 : in  std_logic_vector(STATE_BUS_SIZE - 1 downto 0);
      addr_a0             : in  std_logic_vector(ADDR_BUS_SIZE - 1 downto 0);
      addr_b0             : in  std_logic_vector(ADDR_BUS_SIZE - 1 downto 0);
      addr_c0             : in  std_logic_vector(ADDR_BUS_SIZE - 1 downto 0);
      addr_d0             : in  std_logic_vector(ADDR_BUS_SIZE - 1 downto 0);
      type_data_read_a1   : out std_logic_vector(TYPE_BUS_SIZE - 1 downto 0);
      type_data_read_b1   : out std_logic_vector(TYPE_BUS_SIZE - 1 downto 0);
      type_data_read_c1   : out std_logic_vector(TYPE_BUS_SIZE - 1 downto 0);
      type_data_read_d1   : out std_logic_vector(TYPE_BUS_SIZE - 1 downto 0);
      type_data_write_a1  : in  std_logic_vector(TYPE_BUS_SIZE - 1 downto 0);
      type_data_write_b1  : in  std_logic_vector(TYPE_BUS_SIZE - 1 downto 0);
      type_data_write_c1  : in  std_logic_vector(TYPE_BUS_SIZE - 1 downto 0);
      type_data_write_d1  : in  std_logic_vector(TYPE_BUS_SIZE - 1 downto 0);
      state_data_read_a1  : out std_logic_vector(STATE_BUS_SIZE - 1 downto 0);
      state_data_read_b1  : out std_logic_vector(STATE_BUS_SIZE - 1 downto 0);
      state_data_read_c1  : out std_logic_vector(STATE_BUS_SIZE - 1 downto 0);
      state_data_read_d1  : out std_logic_vector(STATE_BUS_SIZE - 1 downto 0);
      state_data_write_a1 : in  std_logic_vector(STATE_BUS_SIZE - 1 downto 0);
      state_data_write_b1 : in  std_logic_vector(STATE_BUS_SIZE - 1 downto 0);
      state_data_write_c1 : in  std_logic_vector(STATE_BUS_SIZE - 1 downto 0);
      state_data_write_d1 : in  std_logic_vector(STATE_BUS_SIZE - 1 downto 0);
      addr_a1             : in  std_logic_vector(ADDR_BUS_SIZE - 1 downto 0);
      addr_b1             : in  std_logic_vector(ADDR_BUS_SIZE - 1 downto 0);
      addr_c1             : in  std_logic_vector(ADDR_BUS_SIZE - 1 downto 0);
      addr_d1             : in  std_logic_vector(ADDR_BUS_SIZE - 1 downto 0);

      dev_enable_read_a0        : in  std_logic;
      dev_enable_read_b0        : in  std_logic;
      dev_enable_read_c0        : in  std_logic;
      dev_enable_read_d0        : in  std_logic;
      dev_enable_write_a1       : in  std_logic;
      dev_enable_write_c1       : in  std_logic;
      dev_enable_read_b1        : in  std_logic;
      dev_enable_read_d1        : in  std_logic;
      lss_enable_read_a0        : in  std_logic;
      lss_enable_read_c0        : in  std_logic;
      lss_enable_write_type_b0  : in  std_logic;
      lss_enable_write_type_d0  : in  std_logic;
      lss_enable_write_state_b0 : in  std_logic;
      lss_enable_write_state_d0 : in  std_logic;
      stall                     : in  std_logic;
      cfg_enable_read_b1        : in  std_logic;
      cfg_enable_read_d1        : in  std_logic;
      rdb_enable_write_state_a1 : in  std_logic;
      rdb_enable_write_state_b1 : in  std_logic;
      rdb_enable_write_state_c1 : in  std_logic;
      rdb_enable_write_state_d1 : in  std_logic;
      select_sbm                : in  std_logic;
      rst                       : in  std_logic;
      clk                       : in  std_logic);
  end component;

  component hazard
    port (
      dont_issue_dec        : out std_logic;
      stall_lss             : out std_logic;
      stall_dec             : out std_logic;
      stall_fetch           : out std_logic;
      stall_sbm_bram_mgr    : out std_logic;
--Kaa
      stall_usedrules_mem   : out std_logic;
      stall_run_step_mem    : out std_logic;
      stall_rulevector_mem  : out std_logic;
      stall_fitness         : out std_logic;
--Kaa
      fetch_valid           : in  std_logic;
      dec_sbm_pipe_access   : in  std_logic;
      dec_start_devstep     : in  std_logic;
      dec_lss_access        : in  std_logic;
--Kaa
      dec_start_fitness     : in  std_logic;
--Kaa
      lss_idle              : in  std_logic;
      lss_ld2_sending       : in  std_logic;
      lss_ack_send_i        : in  std_logic;
      send                  : in  std_logic;
      dev_idle              : in  std_logic;
      sbm_pipe_idle         : in  std_logic;
--Kaa
      fitness_pipe_idle     : in  std_logic;
--Kaa
      rst                   : in  std_logic;
      clk                   : in  std_logic);
  end component;

  component fetch
    port (
      fetch_instruction       : out std_logic_vector(63 downto 0);
      fetch_valid             : out std_logic;
      fetch_enter_normal_mode : out std_logic;
      fetch_count_pc          : out std_logic;
      program_counter         : in  std_logic_vector(7 downto 0);
      valid                   : in  std_logic;
      program_store           : in  std_logic;
      stall                   : in  std_logic;
      flush                   : in  std_logic;
      receive                 : out std_logic;
      ack_receive             : in  std_logic;
      data_receive            : in  std_logic_vector(31 downto 0);
      rst                     : in  std_logic;
      clk                     : in  std_logic);
  end component;

  component decode
    port (
      fetch_instruction       : in  std_logic_vector(63 downto 0);
      fetch_valid             : in  std_logic;
      fetch_enter_normal_mode : in  std_logic;
      fetch_count_pc          : in  std_logic;
      dec_program_counter     : out std_logic_vector(7 downto 0);
      dec_valid               : out std_logic;
      flush_fetch             : out std_logic;
      dec_program_store       : out std_logic;
      dec_read_sblock         : out std_logic;
--Kaa
      dec_read_usedrules      : out std_logic;
--Kaa
      dec_send_type           : out std_logic;
      dec_send_types          : out std_logic;
      dec_send_state          : out std_logic;
      dec_send_states         : out std_logic;
      dec_send_rulevector     : out std_logic;
      dec_send_fitness        : out std_logic;
      dec_start_fitness       : out std_logic;
--Kaa
      dec_send_sums           : out std_logic;
      dec_send_used_rules     : out std_logic;
--Kaa
      
      dec_clear_bram          : out std_logic;
      dec_write_type          : out std_logic;
      dec_write_state         : out std_logic;
      dec_address             : out std_logic_vector(ADDR_BUS_SIZE - 1 downto 0);
      dec_port_select         : out std_logic;
      dec_sblock_number       : out std_logic;
      dec_type_data           : out std_logic_vector(TYPE_SIZE - 1 downto 0);
      dec_state_data          : out std_logic;


--Kaa
      dec_number_of_readback_values: out
                std_logic_vector (RUN_STEP_ADDR_BUS_SIZE - 1 downto 0); 
--Kaa
      stall                   : in  std_logic;
      dont_issue              : in  std_logic;
      dec_sbm_pipe_access     : out std_logic;
      dec_lss_access          : out std_logic;
      dec_store_rule          : out std_logic;
      dec_rule_number         : out std_logic_vector(RULE_NBR_BUS_SIZE - 1 downto 0);
      dec_rule_to_store       : out std_logic_vector(RULE_SIZE - 1 downto 0);
      dec_nbr_of_last_rule    : out std_logic_vector(7 downto 0);
      dec_lutconv_index       : out std_logic_vector(TYPE_SIZE - 1 downto 0);
      dec_lutconv_write       : out std_logic_vector(LUT_SIZE - 1 downto 0);
      dec_lutconv_write_en    : out std_logic;
      dec_start_devstep       : out std_logic;
      dec_start_config        : out std_logic;
      dec_start_readback      : out std_logic;
      dec_select_sbm          : out std_logic;
      dec_run_matrix          : out std_logic;
      dec_cycles_to_run       : out std_logic_vector(23 downto 0);
      rst                     : in  std_logic;
      clk                     : in  std_logic);
  end component;

  component lss
    port (
      type_data_write_b0  : out std_logic_vector(TYPE_BUS_SIZE - 1 downto 0);
      type_data_write_d0  : out std_logic_vector(TYPE_BUS_SIZE - 1 downto 0);
      state_data_write_b0 : out std_logic_vector(STATE_BUS_SIZE - 1 downto 0);
      state_data_write_d0 : out std_logic_vector(STATE_BUS_SIZE - 1 downto 0);
      addr_a0             : out std_logic_vector(ADDR_BUS_SIZE - 1 downto 0);
      addr_b0             : out std_logic_vector(ADDR_BUS_SIZE - 1 downto 0);
      addr_c0             : out std_logic_vector(ADDR_BUS_SIZE - 1 downto 0);
      addr_d0             : out std_logic_vector(ADDR_BUS_SIZE - 1 downto 0);
      type_data_read_a0   : in  std_logic_vector(TYPE_BUS_SIZE - 1 downto 0);
      type_data_read_c0   : in  std_logic_vector(TYPE_BUS_SIZE - 1 downto 0);
      state_data_read_a0  : in  std_logic_vector(STATE_BUS_SIZE - 1 downto 0);
      state_data_read_c0  : in  std_logic_vector(STATE_BUS_SIZE - 1 downto 0);
--Kaa
      sum_data_read        : in std_logic_vector (15 downto 0);
      sum_address          : out std_logic_vector (RUN_STEP_ADDR_BUS_SIZE - 1 downto 0);
      enable_sum_data_read : out std_logic;

      usedrules_data       :  in std_logic_vector (USEDRULES_DATA_BUS_SIZE - 1 downto 0);
      usedrules_read_enable: out std_logic;
      usedrules_read_addr  : out std_logic_vector (USEDRULES_ADDR_BUS_SIZE - 1 downto 0);

      rulevector_data      : in std_logic_vector(RULEVECTOR_DATA_BUS_SIZE - 1 downto 0);
      read_next_rulevector : out std_logic;
      reset_rulevector_addr: out std_logic;

      fitness_reg_data     : in std_logic_vector(FITNESS_DATA_BUS_SIZE - 1 downto 0);
      fitness_reg_read_enable: out std_logic;
--Kaa

      lss_enable_write_type_b0  : out std_logic;
      lss_enable_write_type_d0  : out std_logic;
      lss_enable_write_state_b0 : out std_logic;
      lss_enable_write_state_d0 : out std_logic;
      lss_enable_read_a0        : out std_logic;
      lss_enable_read_c0        : out std_logic;
      send                      : out std_logic;
      ack_send                  : in  std_logic;
      data_send                 : out std_logic_vector(31 downto 0);
      dec_read_sblock           : in  std_logic;
      dec_send_type             : in  std_logic;
      dec_send_types            : in  std_logic;
      dec_send_state            : in  std_logic;
      dec_send_states           : in  std_logic;
--Kaa
      dec_read_usedrules        : in std_logic;
--Kaa
      dec_send_sums             : in  std_logic;
      dec_send_used_rules       : in  std_logic;
      dec_send_rulevector       : in std_logic;
      dec_send_fitness          : in std_logic;
      dec_number_of_readback_values: in
                 std_logic_vector (RUN_STEP_ADDR_BUS_SIZE - 1 downto 0);
--Kaa
      dec_write_type            : in  std_logic;
      dec_write_state           : in  std_logic;

      dec_address       : in std_logic_vector(ADDR_BUS_SIZE - 1 downto 0);
      dec_port_select   : in std_logic;
      dec_sblock_number : in std_logic;
      dec_type_data     : in std_logic_vector(TYPE_SIZE - 1 downto 0);
      dec_state_data    : in std_logic;

      dec_clear_bram          : in std_logic;

      stall           : in  std_logic;
      lss_idle        : out std_logic;
      lss_ld2_sending : out std_logic;
      lss_ack_send_i  : out std_logic;

      rst      : in  std_logic;
      clk      : in  std_logic);
  end component;

  component dev
    port (
      ruleset             : in  rule_set_t;
      cache_set_zero      : out std_logic;
      cache_next_set      : out std_logic;
      last_set            : in  std_logic;
      type_data_read_a0   : in  std_logic_vector(TYPE_BUS_SIZE - 1 downto 0);
      type_data_read_b0   : in  std_logic_vector(TYPE_BUS_SIZE - 1 downto 0);
      type_data_read_c0   : in  std_logic_vector(TYPE_BUS_SIZE - 1 downto 0);
      type_data_read_d0   : in  std_logic_vector(TYPE_BUS_SIZE - 1 downto 0);
      state_data_read_a0  : in  std_logic_vector(STATE_BUS_SIZE - 1 downto 0);
      state_data_read_b0  : in  std_logic_vector(STATE_BUS_SIZE - 1 downto 0);
      state_data_read_c0  : in  std_logic_vector(STATE_BUS_SIZE - 1 downto 0);
      state_data_read_d0  : in  std_logic_vector(STATE_BUS_SIZE - 1 downto 0);
      addr_a0             : out std_logic_vector(ADDR_BUS_SIZE - 1 downto 0);
      addr_b0             : out std_logic_vector(ADDR_BUS_SIZE - 1 downto 0);
      addr_c0             : out std_logic_vector(ADDR_BUS_SIZE - 1 downto 0);
      addr_d0             : out std_logic_vector(ADDR_BUS_SIZE - 1 downto 0);
      type_data_write_a1  : out std_logic_vector(TYPE_BUS_SIZE - 1 downto 0);
      type_data_write_c1  : out std_logic_vector(TYPE_BUS_SIZE - 1 downto 0);
      type_data_read_b1   : in  std_logic_vector(TYPE_BUS_SIZE - 1 downto 0);
      type_data_read_d1   : in  std_logic_vector(TYPE_BUS_SIZE - 1 downto 0);
      state_data_read_b1  : in  std_logic_vector(STATE_BUS_SIZE - 1 downto 0);
      state_data_read_d1  : in  std_logic_vector(STATE_BUS_SIZE - 1 downto 0);
      state_data_write_a1 : out std_logic_vector(STATE_BUS_SIZE - 1 downto 0);
      state_data_write_c1 : out std_logic_vector(STATE_BUS_SIZE - 1 downto 0);
      addr_a1             : out std_logic_vector(ADDR_BUS_SIZE - 1 downto 0);
      addr_b1             : out std_logic_vector(ADDR_BUS_SIZE - 1 downto 0);
      addr_c1             : out std_logic_vector(ADDR_BUS_SIZE - 1 downto 0);
      addr_d1             : out std_logic_vector(ADDR_BUS_SIZE - 1 downto 0);
      dev_enable_read_a0  : out std_logic;
      dev_enable_read_b0  : out std_logic;
      dev_enable_read_c0  : out std_logic;
      dev_enable_read_d0  : out std_logic;
      dev_enable_write_a1 : out std_logic;
      dev_enable_read_b1  : out std_logic;
      dev_enable_write_c1 : out std_logic;
      dev_enable_read_d1  : out std_logic;
--Kaa
      dev_usedrules_read         :  in std_logic_vector (USEDRULES_DATA_BUS_SIZE - 1 downto 0);
      dev_usedrules_write        : out std_logic_vector (USEDRULES_DATA_BUS_SIZE - 1 downto 0);
      dev_usedrules_addr_read    : out std_logic_vector (USEDRULES_ADDR_BUS_SIZE - 1 downto 0);
      dev_usedrules_addr_write   : out std_logic_vector (USEDRULES_ADDR_BUS_SIZE - 1 downto 0);
      dev_usedrules_read_enable  : out std_logic;
      dev_usedrules_write_enable : out std_logic;

      dev_rulevector_data_write  : out std_logic_vector(RULEVECTOR_DATA_BUS_SIZE - 1 downto 0);
      dev_rulevector_write_enable: out std_logic;
--Kaa
      dec_start_devstep   : in  std_logic;
      dev_idle            : out std_logic;
      rst                 : in  std_logic;
      clk                 : in  std_logic); 
  end component;

  component sbm_pipe
    port (
      state_data_write_a1 : out std_logic_vector(STATE_BUS_SIZE - 1 downto 0);
      state_data_write_b1 : out std_logic_vector(STATE_BUS_SIZE - 1 downto 0);
      state_data_write_c1 : out std_logic_vector(STATE_BUS_SIZE - 1 downto 0);
      state_data_write_d1 : out std_logic_vector(STATE_BUS_SIZE - 1 downto 0);
      type_data_read_b1   : in  std_logic_vector(TYPE_BUS_SIZE - 1 downto 0);
      type_data_read_d1   : in  std_logic_vector(TYPE_BUS_SIZE - 1 downto 0);
      state_data_read_b1  : in  std_logic_vector(STATE_BUS_SIZE - 1 downto 0);
      state_data_read_d1  : in  std_logic_vector(STATE_BUS_SIZE - 1 downto 0);
      addr_a1             : out std_logic_vector(ADDR_BUS_SIZE - 1 downto 0);
      addr_b1             : out std_logic_vector(ADDR_BUS_SIZE - 1 downto 0);
      addr_c1             : out std_logic_vector(ADDR_BUS_SIZE - 1 downto 0);
      addr_d1             : out std_logic_vector(ADDR_BUS_SIZE - 1 downto 0);

      cfg_enable_read_b1        : out std_logic;
      cfg_enable_read_d1        : out std_logic;
      rdb_enable_write_state_a1 : out std_logic;
      rdb_enable_write_state_b1 : out std_logic;
      rdb_enable_write_state_c1 : out std_logic;
      rdb_enable_write_state_d1 : out std_logic;

      index_a              : out std_logic_vector(TYPE_SIZE - 1 downto 0);
      index_b              : out std_logic_vector(TYPE_SIZE - 1 downto 0);
      lut_a_read           : in  std_logic_vector(LUT_SIZE - 1 downto 0);
      lut_b_read           : in  std_logic_vector(LUT_SIZE - 1 downto 0);
      databus_lut_l_write  : out std_logic_vector(SBM_CFG_SIZE - 1 downto 0);
      databus_lut_h_write  : out std_logic_vector(SBM_CFG_SIZE - 1 downto 0);
      databus_ff_write     : out std_logic_vector(SBM_CFG_SIZE - 1 downto 0);
      config_enable_lut    : out std_logic_vector(CONFIG_WORDS - 1 downto 0);
      config_enable_ff     : out std_logic_vector(CONFIG_WORDS - 1 downto 0);
      databus_read         : in  std_logic_vector(SBM_RDB_SIZE - 1 downto 0);
      output_select        : out std_logic_vector(READBACK_WORDS - 1 downto 0);
      dec_start_config     : in  std_logic;
      dec_start_readback   : in  std_logic;
      dec_run_matrix       : in  std_logic;
      dec_cycles_to_run    : in  std_logic_vector(23 downto 0);
      sbm_pipe_idle        : out std_logic;
      run_matrix           : out std_logic;
      --Kaa
      add                  : out std_logic;
      clear_sum            : out std_logic;

      run_step_mem_address :out std_logic_vector (RUN_STEP_ADDR_BUS_SIZE - 1 downto 0);
      run_step_mem_write_enable : out std_logic;
      --Kaa
      rst                  : in  std_logic;
      clk                  : in  std_logic);
  end component;

--Kaa
--   component sum_modul
--     port (
--       data_bus : in std_logic_vector (SBM_RDB_SIZE - 1  downto 0);
--       add      : in std_logic;
--       sum      : out std_logic_vector (RUN_STEP_DATA_BUS_SIZE - 1 downto 0);
--       clear_sum: in std_logic;
--       clk      : in std_logic;
--       rst      : in std_logic);
--   end component;

  component run_step_funk
    port (
    data_bus : in  std_logic_vector (SBM_RDB_SIZE - 1 downto 0);
    active   : in  std_logic;
    clear    : in  std_logic;
    address_in     : in std_logic_vector (RUN_STEP_ADDR_BUS_SIZE - 1 downto 0);
    write_enable_in: in std_logic;
    value           : out std_logic_vector (RUN_STEP_DATA_BUS_SIZE - 1 downto 0);
    address_out     : out std_logic_vector (RUN_STEP_ADDR_BUS_SIZE - 1 downto 0);
    write_enable_out: out std_logic;
    clk      : in  std_logic;
    rst      : in  std_logic);
  end component;


  component bitcounter4
    port (
      a : in  std_logic;
      b : in  std_logic;
      c : in  std_logic;
      d : in  std_logic;
      f : out std_logic_vector (2 downto 0));
  end component;

  component bitcounter8 
    port (
      input : in  std_logic_vector (7 downto 0);  --between 0 and 4
      f : out std_logic_vector (3 downto 0));
  end component;

  component bitcounter16
    port (
      input : in  std_logic_vector (15 downto 0);  --between 0 and 16
      f : out std_logic_vector (4 downto 0));
  end component;

  component run_step_mem
    port (
      address      : in  std_logic_vector (RUN_STEP_ADDR_BUS_SIZE - 1 downto 0);
      data_read    : out std_logic_vector (RUN_STEP_DATA_BUS_SIZE - 1 downto 0);
      data_write   : in  std_logic_vector (RUN_STEP_DATA_BUS_SIZE - 1 downto 0);
      write_enable : in  std_logic;
      read_enable  : in  std_logic;   

      stall        : in std_logic;
      rst          : in std_logic;
      clk          : in std_logic);
  end component;
  
  component rulevector_mem
    port (
      data_read    : out std_logic_vector (RULEVECTOR_DATA_BUS_SIZE - 1 downto 0);
      data_write   :  in std_logic_vector (RULEVECTOR_DATA_BUS_SIZE - 1 downto 0);
      write_enable :  in std_logic;
      reset_counter:  in std_logic;
      read_next    :  in std_logic;
      
      stall        :  in std_logic;
      
      rst : in std_logic;
      clk : in std_logic);
  end component;

  component usedrules_mem
    port (
      address_read : in  std_logic_vector (USEDRULES_ADDR_BUS_SIZE - 1 downto 0);--512 addresses
      address_write: in  std_logic_vector (USEDRULES_ADDR_BUS_SIZE - 1 downto 0);--512 addresses
      data_read    : out std_logic_vector (USEDRULES_DATA_BUS_SIZE - 1 downto 0);
      data_write   : in  std_logic_vector (USEDRULES_DATA_BUS_SIZE - 1 downto 0);
      write_enable : in  std_logic;
      --read_enable  : in  std_logic;   
      stall        : in std_logic;
      rst          : in std_logic;
      clk          : in std_logic);
  end component;

  component decoder_8to256
    port (
      input        : in std_logic_vector(  7 downto 0);
      output       :out std_logic_vector(255 downto 0));
  end component;

  component fitness_pipe
    port (
      dec_start_fitness   :  in std_logic;
      fitness_data        : out std_logic_vector(FITNESS_DATA_BUS_SIZE - 1 downto 0);
      read_enable         :  in std_logic;
      data_in             :  in std_logic_vector(RUN_STEP_DATA_BUS_SIZE - 1 downto 0);
      data_addr           : out std_logic_vector(RUN_STEP_ADDR_BUS_SIZE - 1 downto 0);
      run_step_to_evaluate:  in std_logic_vector(RUN_STEP_ADDR_BUS_SIZE - 1 downto 0);
      -- hazard
      stall               :  in std_logic;
      fitness_idle        : out std_logic;
      -- other
      rst                 : in std_logic;
      clk                 : in std_logic);
  end component;

  component fitness_funk
    port (
      ld_finished      :  in std_logic;
      fitness_finished : out std_logic;
      active           :  in std_logic;
      data             :  in std_logic_vector(RUN_STEP_DATA_BUS_SIZE - 1 downto 0);
      result           : out std_logic_vector(FITNESS_RESULT_SIZE - 1 downto 0);
      stall            :  in std_logic;
      rst              :  in std_logic;
      clk              :  in std_logic);

  end component;
--Kaa

end sblock_package;

