/* sblocklib.c
 *
 * Sblocklib is an abstraction layer above FUSE API for controlling the
 * coprosessor 
 * 
 * Asbjorn Djupdal 2003
 * Kjetil Aamodt 2005
 */

/* For information about functions in this file, see comments
   in sblocklib.h */

#include "sblocklib.h"
#include <benera.h>

#define PCI_CLK    2
#define PCI_FREQ  40

#define CLK     1
#define FREQ   80

static LOCATE_HANDLE hLocate = NULL;
static DIME_HANDLE hCard;
static DIME_MEMHANDLE hMemSend;
static DIME_MEMHANDLE hMemReceive;
static DIME_DMAHANDLE hDMA;
static int bufferPtr;
static uint32_t sendBuffer[BUFFER_SIZE];
uint32_t receiveBuffer[BUFFER_SIZE];

/*****************************************************************************/
/* functions related to card control */

bool_t openCard (int number) {
  bool_t success = FALSE;

  /* locate card */
	if ((hLocate = DIME_LocateCard (dlPCI, mbtALL, NULL, dldrDEFAULT,
                                  dlDEFAULT))) {
    /* open card */
		if ((hCard=DIME_OpenCard (hLocate, number, dccOPEN_DEFAULT))) {
      /* lock memory used for DMA buffers */
      if ((hMemReceive = DIME_LockMemory (hCard, (DWORD*)receiveBuffer,
                                   sizeof receiveBuffer))) {
        if ((hMemSend = DIME_LockMemory (hCard, (DWORD*)sendBuffer,
                                     sizeof sendBuffer))) {
          bufferPtr = 0;

          /* get dma channel */
          if ((hDMA = DIME_DMAOpen (hCard, 1, 0))) {
            /* set max jtag speed */
            if (!DIME_JTAGControl (hCard, djtagCONFIGSPEED,
                                   djtagMAXSPEED100)) {
              /* set clock speed */
              DIME_SetOscillatorFrequency (hCard, PCI_CLK, PCI_FREQ, NULL);
              DIME_SetOscillatorFrequency (hCard, CLK, FREQ, NULL);

              success = TRUE;
            }
          }
        }
      }
    }
  }
  if (!success) {
    closeCard();
  }
  return success;
}

void closeCard (void) {
  /* close card */
  if (hCard) DIME_CloseCard (hCard);
  /* close locate */
  if (hLocate) DIME_CloseLocate (hLocate);
}

bool_t configureCard (char* bitfilename) {
  /* konfigurasjon */
  DIME_ConfigSetBitsFilenameAndConfig (hCard, 0, 2, bitfilename, 0, NULL, 0);
  /* reset */
  DIME_CardResetControl (hCard, drONBOARDFPGA, drENABLE, 0);
  DIME_CardResetControl (hCard, drINTERFACE, drTOGGLE, 0);
  DIME_CardResetControl (hCard, drONBOARDFPGA, drDISABLE, 0);

  return TRUE;
}

void flushDMA (void) {
  DIME_DMAWriteFromLockedMem (hCard, hDMA, hMemSend, 0, bufferPtr, 0);
  bufferPtr = 0;
}

void insertDMA (uint32_t data) {
  sendBuffer[bufferPtr++] = data;

  if (bufferPtr == BUFFER_SIZE) {
    flushDMA();
  }
}

void readDMA (int words) {
  DIME_DMAReadToLockedMem (hCard, hDMA, hMemReceive, 0, words, 0);
}

/*****************************************************************************/
/* instructions */

//Kaa
void readSums(uint32_t numberOfReadbacks){
  insertDMA(0x00000014 | (numberOfReadbacks << 8));
}

void readUsedRules(void){
  insertDMA(0x00000015);
}

void resetDevCounter(void){
  insertDMA(0x00000017);
}

void jumpEqual(uint_t value, uint32_t addr){
  insertDMA(0x00000016 | (value << 8) | (addr << 24));
} 

void readRuleVector(uint32_t number){
  insertDMA(0x00000018 | (number << 8));
}

void readFitness(void){
  insertDMA(0x00000019);
}

void doFitness(uint32_t number){
  insertDMA(0x0000001A | (number << 8));
}

//Kaa

void store (uint_t addr) {
  insertDMA (0x0000000e | (addr << 8));
}  

void end (void) {
  insertDMA (0x0000000f);
}

inline void nop (void) {
  insertDMA (0x00000000);
}

void jump (uint_t addr) {
  insertDMA (0x0000000c | (addr << 8));
}

void break_prg (void) {
  insertDMA (0x0000000d);
}

void switchSBMs (void) {
  insertDMA (0x00000003);
}

void run (int cycles) {
  insertDMA (0x00000009 | (cycles << 8));
}  

void writeLUTConv (uint32_t lut, int number) {
  insertDMA (0x00000026 | (number << 8));
  insertDMA (lut);
}

void config (void) {
  insertDMA (0x00000007);
}

void readback (void) {
  insertDMA (0x00000008);
}

void devstep (void) {
  insertDMA (0x0000000a);
}

void writeRule (struct RuleStruct rule, int number) {
  uint64_t bin_rule = 0L;

  bin_rule |= (uint64_t)(rule.valid ? 1 : 0) << 48;
  bin_rule |= (uint64_t)(rule.type  ? 1 : 0) << 47;

  bin_rule |= (uint64_t)(rule.north.state_dc     ? 1 : 0) << 46;
  bin_rule |= (uint64_t)(rule.north.state_value  ? 1 : 0) << 45;
  bin_rule |= (uint64_t)(rule.north.type_dc      ? 1 : 0) << 44;
  bin_rule |= (uint64_t)(rule.north.type_value   & 0x1f)  << 39;

  bin_rule |= (uint64_t)(rule.south.state_dc     ? 1 : 0) << 38;
  bin_rule |= (uint64_t)(rule.south.state_value  ? 1 : 0) << 37;
  bin_rule |= (uint64_t)(rule.south.type_dc      ? 1 : 0) << 36;
  bin_rule |= (uint64_t)(rule.south.type_value   & 0x1f)  << 31;

  bin_rule |= (uint64_t)(rule.east.state_dc      ? 1 : 0) << 30;
  bin_rule |= (uint64_t)(rule.east.state_value   ? 1 : 0) << 29;
  bin_rule |= (uint64_t)(rule.east.type_dc       ? 1 : 0) << 28;
  bin_rule |= (uint64_t)(rule.east.type_value    & 0x1f)  << 23;

  bin_rule |= (uint64_t)(rule.west.state_dc      ? 1 : 0) << 22;
  bin_rule |= (uint64_t)(rule.west.state_value   ? 1 : 0) << 21;
  bin_rule |= (uint64_t)(rule.west.type_dc       ? 1 : 0) << 20;
  bin_rule |= (uint64_t)(rule.west.type_value    & 0x1f)  << 15;

  bin_rule |= (uint64_t)(rule.center.state_dc    ? 1 : 0) << 14;
  bin_rule |= (uint64_t)(rule.center.state_value ? 1 : 0) << 13;
  bin_rule |= (uint64_t)(rule.center.type_dc     ? 1 : 0) << 12;
  bin_rule |= (uint64_t)(rule.center.type_value  & 0x1f)  << 7;

  bin_rule |= (uint64_t)(rule.dont_change_state ? 1 : 0) << 6;

  if (rule.type == CHANGE) {
    bin_rule |= (uint64_t)(rule.result.change.state ? 1 : 0) << 5;
    bin_rule |= (uint64_t)(rule.result.change.type & 0x1f)   << 0;
  } else {
    bin_rule |= (uint64_t)(rule.result.copyFrom & 3) << 0;
  }

  insertDMA (0x0000002b | ((number & 0xff) << 7) |
             ((bin_rule & 0x1ffff) << 15));
  insertDMA ((bin_rule >> 17L) & 0xffffffff);
}

void writeState (bool_t state, int x, int y) {
  insertDMA (0x00000004 | (state ? (1 << 31) : 0) | (y << 16) | (x << 8));
}

void readState (int x, int y) {
  insertDMA (0x00000005 | (y << 16) | (x << 8));
}

void writeType (int type, int x, int y) {
  insertDMA (0x00000001 | (type << 24) | (y << 16) | (x << 8));
}

void readType (int x, int y) {
  insertDMA (0x00000002 | (y << 16) | (x << 8));
}

void readStates (void) {
  insertDMA (0x00000011);
}

void readTypes (void) {
  insertDMA (0x00000010);
}

void setNumberOfLastRule (int numberOfLastRule) {
  insertDMA (0x00000012 | (numberOfLastRule << 8));
}

void clearBRAM (int type, bool_t state) {
  insertDMA (0x00000013 | (type << 24) | (state ? (1 << 31) : 0));
}

/*****************************************************************************/
/* utility functions */

void saveSendBuffer (char* modelsim) {
  int i;
  FILE* modelsimfile;

  if(modelsim[0] != ' '){ //save buffer only if filename is given
    if ((modelsimfile = fopen (modelsim, "wb"))) {
      for (i = 0; i < bufferPtr; i++) {
	fprintf (modelsimfile, "force -freeze sim:/toplevel/pciempty 0 0\n"
		 "run 25ns\n"
		 "force -drive sim:/toplevel/pcidata 16#%08X 0\n"
		 "run 25ns\n"
		 "force -freeze sim:/toplevel/pciempty 1 0\n"
		 "noforce sim:/toplevel/pcidata\n"
		 "run 200ns\n\n",
		 (uint32_t)sendBuffer[i]);
      }
      fclose (modelsimfile);
    }
    else{
      printf("Can't open modelsimfile\n");
    }
  }
}
