/* sblocktest.c
 *
 * A testprogram for sblockmatrix project
 *
 * This runs same test experiments that is described in the master
 * thesis report.  
 *
 * Asbjorn Djupdal 2003
 * Kjetil Aamodt 2005
 */

#include "sblocklib.h"
#include "read_print.h"
#include "types.h"
#include "rules.h"

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <unistd.h>
#include <sys/time.h>

/*****************************************************************************/
/* protos */

/* write some specific rules to rule storage */
void writeRule_empty (int number);
void writeRule_change_a (int number);
void writeRule_change_b (int number);
void writeRule_grow_north (int number);
void writeRule_grow_south (int number);
void writeRule_grow_east (int number);
void writeRule_grow_west (int number);

/* write a specific state pattern to BRAM-0 */
void setStatesAlternating (void);

/* start the co-processor */
void startCP(void);

/* write program to co-processor */
void store_program_test_funct(void);
void store_program_test_rule(void);
void store_program_test_speed(void);
void store_program_test_fitness(void);
void store_program_test_fitness_speed_new(int runstep);


/* read from PCI and write to file */
void readback_test_funct(void);
void readback_test_rule(void);
void readback_test_speed(void);
void readback_test_fitness(void);
void readback_test_fitness_speed_new(int devstep);

void test_fitness_speed_old();

/*****************************************************************************/

static FILE* outputfile;

enum test_t {TEST_FUNCT = 0, TEST_SPEED, TEST_RULE, TEST_FITNESS, 
	     TEST_FITNESS_SPEED_OLD, TEST_FITNESS_SPEED_NEW};
enum test_t test;

int card;
char *modelsim;
char *bitfile;
char *output;

/*****************************************************************************/
/* functions */

int main (int argc, char* argv[]) {
  card = 1; //default
  test = 0;
  modelsim = " ";
  bitfile = argv[1];
  output = argv[2];

  //printf("Start with %i parameters\n", argc);
  if(argc > 2){
    if (argc > 3){ //testnumber is given
      test = atoi(argv[3]);
      if (argc > 4){ //cardnumber is given
	card = atoi(argv[4]);
	if (argc > 5){ //modelsimfile is given
	  modelsim = argv[5];
	}
      }
    }

    printf("Bitfile: \t%s\nOutputfile: \t%s\nTest: \t\t%i\nCard: \t\t%i\nModelsim: \t%s\n\n", 
	   bitfile, output, test, card, modelsim);
    fflush(stdout);
    
    if(openCard (card)) {
      if (configureCard (bitfile)) {
	if ((outputfile = fopen (output, "wb"))) {
	  /* get current time, used to measure total time used
	     (exclusive startup time) at the end */
	  struct timeval tv_start;
	  struct timeval tv_end;
	  gettimeofday (&tv_start, NULL);
	  
	  /* write to lut conv table */
	  writeLUTConv (EMPTY, 0); /* sblock type 0 is empty */
	  writeLUTConv (XOR4, 1);  /* sblock type 1 is XOR4 */
	  writeLUTConv (OR4, 2);   /* sblock type 2 is OR4 */ 
	  writeLUTConv (AND4, 3);  /* sblock type 3 is AND4 */
	  
	  /* write rules to rule storage */
	  writeRule(create_rule_grow_south(), 0);
	  writeRule(create_rule_grow_east(), 1);
	  writeRule(create_rule_grow_north(), 2);
	  writeRule(create_rule_grow_west(), 3);
	  writeRule(create_rule_change_a(), 4);
	  writeRule(create_rule_change_b(), 5);
	  
	  setNumberOfLastRule  (5);
  
	  clearBRAM(0, 0);
	  
	  switch (test){
	  case TEST_FUNCT:
	    printf("Test Function\n");
	    store_program_test_funct();
	    saveSendBuffer (modelsim);
	    startCP();
	    readback_test_funct();
	    break;
	  case TEST_SPEED:
	    printf("Test Speed\n");
	    store_program_test_speed();
	    saveSendBuffer (modelsim);
	    startCP();
	    readback_test_speed();
	    break;
	  case TEST_RULE:
	    printf("Test Rule\n");
	    store_program_test_rule();
	    saveSendBuffer (modelsim);
	    startCP();
	    readback_test_rule();
	    break;
	  case TEST_FITNESS:
	    printf("Test Fitness\n");
	    store_program_test_fitness();
	    saveSendBuffer(modelsim);
	    startCP();
	    readback_test_fitness();
	    break;
	  case TEST_FITNESS_SPEED_OLD:
	    printf("Test Fitness speed old\n");
	    test_fitness_speed_old(1000, 1000);
	    break;
	  case TEST_FITNESS_SPEED_NEW:
	    printf("Test Fitness speed new\n");
	    store_program_test_fitness_speed_new(1000);
	    saveSendBuffer(modelsim);
	    startCP();
	    readback_test_fitness_speed_new(1000);
	    break;
	  default:
	    break;
	  }
	
	  /* get current time and print out total time used */
	  gettimeofday (&tv_end, NULL);
	  printf ("%ld.%06lds\n",
		  tv_end.tv_sec - tv_start.tv_sec -
		  (tv_end.tv_usec < tv_start.tv_usec ? 1 : 0),
		  
		  tv_end.tv_usec - tv_start.tv_usec +
		  (tv_end.tv_usec < tv_start.tv_usec ? 1000000 : 0));
	  
	  fclose (outputfile);
	} else fprintf (stderr, "can't open outputfile\n");
      } else fprintf (stderr, "can't configure with %s\n", argv[1]);
      closeCard();
    } else fprintf (stderr, "can't open card\n");
  } else fprintf (stderr, "Arguments: <bitfile> <outputfile> [[[<test>]"
		  " <card>] <modelsim scriptfile>]\n");
  return 0;
}

/* start coprosessor */
void startCP(void){
  jump (PROGRAM_ADDRESS);
  flushDMA();
}

void setStatesAlternating (void) {
  int i, j;
  for (i = 0; i < ROWS; i++) {
    for (j = 0; j < (COLUMNS / 2); j++) {
      writeState (1, j * 2 + 1, i);
    }
  }
}

void test_fitness_speed_old(int runstepcounter, int devstepcounter){
  int i, j;
  setStatesAlternating();
  writeType (2, 4, 1);
  switchSBMs();

  store(PROGRAM_ADDRESS);{
      run(1);
      readback();
      switchSBMs();
      readStates();
      switchSBMs();
      config();
      break_prg();
  }end();
  flushDMA();

  for(j=0;j<devstepcounter;j++){
    config(); flushDMA();
    for(i=0;i<runstepcounter;i++){
      fprintf(outputfile, "run-step\n");

      jump(PROGRAM_ADDRESS); flushDMA();

      readDMAStates();
      printStates(outputfile);
    }
    readback();
    switchSBMs();
    devstep();
    flushDMA();
    fprintf(outputfile, "Development\n"); fflush(stdout);
  }
}

void store_program_test_rule(){
  switchSBMs();

  store(PROGRAM_ADDRESS);{
    config();
    readback();
    switchSBMs();
    readTypes();
    devstep();
    readUsedRules();
    
    jumpEqual(20, PROGRAM_ADDRESS + 30);

    jump(PROGRAM_ADDRESS);
  }end();

  store(PROGRAM_ADDRESS + 30);{
    readRuleVector(20);
    jump(PROGRAM_ADDRESS);
  }end();

}

void readback_test_rule(void){
  int i;

  for (i = 0; i < 20; i++) {
    fprintf (outputfile, "%d\n", i);

    readDMATypes();
    printTypes(outputfile);

    readDMAUsedRules();
    printUsedRules(outputfile);
  }
  
  readDMAVector(20);
  printVector(20, outputfile);
}

void store_program_test_funct(){
  /* set states */
  setStatesAlternating();
  
  /* set types */
  writeType (2, 4, 1);

  /* change so that initial data is in BRAM-1 */
  switchSBMs();
  
  /* store program */
  store (PROGRAM_ADDRESS); {
    config();
    run (77);
    readback();
    
    switchSBMs();

    readTypes();
    readStates();

    devstep();
    jump (PROGRAM_ADDRESS);
  }end();
}

void readback_test_funct(void){
  int i;

  for (i = 0; i < 150; i++) {
    fprintf (outputfile, "%d\n", i);

    readDMATypes();
    printTypes(outputfile);

    readDMAStates();
    printStates(outputfile);
  }
}


void store_program_test_speed(){
  /* set states */
  setStatesAlternating();
  
  /* set types */
  writeType (2, 4, 1);

  /* change so that initial data is in BRAM-1 */
  switchSBMs();
  
  /* store program */
  store (PROGRAM_ADDRESS); {
    config();
    run (50000);
    readback();
    
    switchSBMs();

    readTypes();
    readStates();

    devstep();

    jump (PROGRAM_ADDRESS);
  }end();
}

void readback_test_speed(void){
  int i;

  for (i = 0; i < 10000; i++) {
    fprintf (outputfile, "%d\n", i);

    readDMATypes();
    printTypes(outputfile);

    readDMAStates();
    printStates(outputfile);
  }
}

void store_program_test_fitness(void){
  /* set states */
  setStatesAlternating();
  /* set types */
  writeType (2, 4, 1);
  /* change so that initial data is in BRAM-1 */
  switchSBMs();
  
  /* store program */
  store (PROGRAM_ADDRESS); {
    config();
    run (100);
    readback();
    
    switchSBMs();

    readStates();
    readSums(100);

    doFitness(100);
    readFitness();
    readFitness();
    readFitness();
    readFitness();
    devstep();

    jump (PROGRAM_ADDRESS);
  }end();
}

void readback_test_fitness(void){
  int i;
  for(i = 0; i < 100; i++){
    fprintf (outputfile, "%d\n", i);
    
    readDMAStates();
    printStates(outputfile);
    
    readDMASums(100);
    printSums(100, outputfile);

    readAndPrintFitness(outputfile);
    readAndPrintFitness(outputfile);
    readAndPrintFitness(outputfile);
    readAndPrintFitness(outputfile);
  }
}

void store_program_test_fitness_speed_new(int runstep){
  /* set states */
  setStatesAlternating();
  /* set types */
  writeType (2, 4, 1);
  /* change so that initial data is in BRAM-1 */
  switchSBMs();
  
  /* store program */
  store (PROGRAM_ADDRESS); {
    config();
    run (runstep);
    readback();
    
    switchSBMs();

    doFitness(runstep);
    readFitness();
    readFitness();
    readFitness();
    readFitness();
    devstep();

    jump (PROGRAM_ADDRESS);
  }end();
}

void readback_test_fitness_speed_new(int devstep){
  int i;
  for(i = 0; i < devstep; i++){
    fprintf (outputfile, "%d\n", i);
    readAndPrintFitness(outputfile);
    readAndPrintFitness(outputfile);
    readAndPrintFitness(outputfile);
    readAndPrintFitness(outputfile);
  }
}

